import sys,os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import cv2
from recover import recover_vel, linear_reconstruction
import argparse

def pltimg(img, title):
  if len(img.shape)>2:
    img = img[0]
  plt.figure()
  plt.imshow(np.real(img))
  plt.colorbar()
  plt.title(title)
  plt.draw()

def save_mask(maskFile, outputFile):
  '''
  Save undersampling mask to file
  '''
  mask = np.load(maskFile).astype(bool)
  if(len(mask.shape) == 3):
    plt.imshow(np.absolute(np.fft.fftshift(mask[0])), cmap='gray', vmin=0, vmax=1)
  else:
    plt.imshow(np.absolute(np.fft.fftshift(mask)), cmap='gray', vmin=0, vmax=1)
  plt.axis('off')
  plt.savefig(outputFile, bbox_inches='tight', pad_inches=0)
  plt.close()

def save_rec(infilename, venc, p, samptype, noise_percent, wavetype, algtype, prefix, maindir, outputdir, limits, fluidmaskfile, singleChannel=False, relative=False):
  
  recovered = np.load(infilename)

  if(singleChannel):
    imgs = np.absolute(recovered)
  else:
    imgs = recover_vel(recovered, venc)

  # Set to zero the pixels outside the fluid mask
  if(fluidmaskfile != ''):
    print('Using fluid mask: ',maindir + fluidmaskfile)
    fluidmask = np.load(maindir + fluidmaskfile)
    imgs[0,:,0,np.logical_not(fluidmask)] = 0.0
    
  # Only the first reconstructed Sample
  for n in range(1): #range(imgs.shape[0]):
    for k in range(imgs.shape[1]):
      # Custom image Scaling for Comparison
      myvmin=limits[2*k]
      myvmax=limits[2*k+1]
      if(myvmin is not None):
        print('-- Using min value: ',myvmin)
      if(myvmax is not None):
        print('-- Using max value: ',myvmax)

      # Show pictures
      plt.imshow(imgs[n,k,0], cmap='gray',vmin=myvmin,vmax=myvmax)
      plt.axis('off')
      plt.savefig(outputdir + prefix + 'rec_p' + str(int(p*100)) + samptype + \
                                      '_noise' + str(int(noise_percent*100)) + \
                                          '_n' + str(n) + \
                                          '_w' + str(wavetype) + \
                                          '_a' + str(algtype) + \
                                          '_k' + str(k) + '.png', bbox_inches='tight', pad_inches=0)
      plt.close()

def save_lin(infilename,
             venc, p, samptype, noise_percent, wavetype, num_samples,
             prefix, maindir, outputdir, limits, fluidmaskfile, usemultipatterns,
             singleChannel=False, relative=False):

  if(usemultipatterns):
    mask_file = maindir + 'undersamplpattern_p' + str(int(p*100)) + samptype + '_n' + str(num_samples) + '.npy'
  else:
    mask_file = maindir + 'undersamplpattern_p' + str(int(p*100)) + samptype + '.npy'

  mask = np.load(maindir + mask_file)
  noisy = np.load(maindir + 'noisy_noise' + str(int(noise_percent*100)) + '_n' + str(num_samples) + '.npy')
  linrec = linear_reconstruction(noisy,mask)

  if(singleChannel):
    imgs = np.absolute(linrec)
  else:
    imgs = recover_vel(linrec, venc)

  # Set to zero the pixels outside the fluid mask
  if(fluidmaskfile != ''):
    print('Using fluid mask: ',maindir + fluidmaskfile)
    fluidmask = np.load(maindir + fluidmaskfile)
    imgs[0,:,0,np.logical_not(fluidmask)] = 0.0
    
  # Only the first reconstructed Sample
  for n in range(1): #range(imgs.shape[0]):
    for k in range(imgs.shape[1]):
      # Custom image Scaling for Comparison
      myvmin=limits[2*k]
      myvmax=limits[2*k+1]
      if(myvmin is not None):
        print('-- Using min value: ',myvmin)
      if(myvmax is not None):
        print('-- Using max value: ',myvmax)

      # Show pictures
      plt.imshow(imgs[n,k,0], cmap='gray',vmin=myvmin,vmax=myvmax)
      plt.axis('off')
      plt.savefig(outputdir + prefix + 'linrec_p' + str(int(p*100)) + samptype + \
                                      '_noise' + str(int(noise_percent*100)) + \
                                          '_n' + str(n) + \
                                          '_w' + str(wavetype) + \
                                          '_k' + str(k) + '.png', bbox_inches='tight', pad_inches=0)
      plt.close()


def save_rec_noise(recnpyFile, orig_file, venc, p, samptype, noise_percent, wavetype, algtype, prefix, outputdir, singleChannel=False, use_truth=False, ext='.png'):
  
  # Read File with CS Reconstructions 
  recovered = np.load(recnpyFile)
  
  # Read True underlying Image
  true = np.load(orig_file)

  # Check consistency 
  if(recovered.shape[-2:] != true.shape[-2:]):
    print('ERROR: reconstructed images and true images are not compatible')
    sys.exit(-1)

  # Reconstruct velocities from complex recovered images
  if(singleChannel):
    imgs = np.absolute(recovered)
  else:
    imgs = recover_vel(recovered, venc)

  # Compute Average Image
  if(not(use_truth)):
    avg = imgs.mean(axis=0)

  # # DEBUG - CHECK THE DIFFERENCE IN RECONSTRUCTED IMAGE
  # plt.figure()
  # plt.subplot(1,2,1)
  # plt.title('TRUE')
  # plt.imshow(np.absolute(true[0,3,0]))
  # plt.colorbar()
  # plt.subplot(1,2,2)
  # plt.title('REC')
  # plt.imshow(np.absolute(imgs[0,3,0]))
  # plt.colorbar()
  # plt.show()
  # exit(-1)

  # Substract images either from the truth or the average reconstruction
  imgRes = np.zeros_like(imgs)
  for i in range(imgs.shape[0]):
    if use_truth:
      boolMask = np.logical_or((np.absolute(imgs[i,:,:,:,:]) < 1.0e-12),(np.absolute(true[0,:,:,:,:]) < 1.0e-12))
      with np.errstate(divide='ignore', invalid='ignore'):
        imgRes[i,:,:,:,:] = 2*(imgs[i,:,:,:,:] - true[0,:,:,:,:])/(np.absolute(imgs[i,:,:,:,:]) + np.absolute(true[0,:,:,:,:]))      
      imgRes[i,boolMask] = 0.0
    else:
      boolMask = np.logical_or((np.absolute(imgs[i,:,:,:,:]) < 1.0e-12),(np.absolute(true[0,:,:,:,:]) < 1.0e-12))
      with np.errstate(divide='ignore', invalid='ignore'):
        imgRes[i,:,:,:,:] = 2*(imgs[i,:,:,:,:] - avg[:,:,:,:])/(np.absolute(imgs[i,:,:,:,:]) + np.absolute(avg[:,:,:,:]))
      imgRes[i,boolMask] = 0.0
      
  if(use_truth):
    desc = 'true'
  else:
    desc = 'avg'

  for n in range(1): # range(imgs.shape[0]): # Loop on the number of samples
    for k in range(imgRes.shape[1]):
      plt.imshow(imgRes[n,k,0], cmap='seismic', vmin=-2, vmax=2)
      plt.axis('off')
      # plt.show()
      plt.savefig(outputdir + prefix + desc + 'recerror' + '_p' + str(int(p*100)) + samptype + \
                                                       '_noise' + str(int(noise_percent*100)) + \
                                                           '_n' + str(n) + \
                                                           '_w' + str(wavetype) + \
                                                           '_a' + str(algtype) + \
                                                           '_k' + str(k) + ext, bbox_inches='tight', pad_inches=0)
      plt.close()
      
def save_noisy(noise_percent, dir, recdir, venc, num_samples, relative=False, ext='.png'):
    noisy = np.load(dir + 'noisy_noise' + str(int(noise_percent*100)) + '_n' + str(num_samples) + '.npy')
    noisy = linear_reconstruction(noisy)
    avg = noisy.mean(axis=0)
    orig_file = dir+'imgs_n1' +  '.npy'
    imgs = recover_vel(noisy, venc)
    # if relative:
    #     imgs = rescale(imgs)
    directory = recdir + 'noise' + str(int(noise_percent*100))
    if not os.path.exists(directory):
        os.makedirs(directory)
    for n in range(5):#range(imgs.shape[0]):
        for k in range(imgs.shape[1]):    
            cv2.imwrite(directory + '/noisy' + '_n' + str(n) + '_k' + str(k) + ext, imgs[n,k,0])

def save_true(trueFileName, outputDir, relative=False):
  true = np.load(trueFileName)
  # if relative:
  #  true = rescale(true, truncate=False)
  for k in range(true.shape[1]):
    if(False):
      print('Min True: ',np.min(true[0,k,0]))
      print('Max True: ',np.max(true[0,k,0]))
    plt.imshow(true[0,k,0], cmap='gray')
    plt.axis('off')
    plt.savefig(outputDir + 'true' '_k' + str(k) + '.png', bbox_inches='tight', pad_inches=0)
    

def save_all(args,relativeScale=False):
  '''
  Save all the pictures determining a reconstruction process
  '''

  # Store Image Prefix
  prefstr = args.imgprefix + '_'  

  # Save Original Image
  if(args.savetrue):
    trueFileName = args.maindir+'imgs_n1.npy'    
    if(os.path.exists(trueFileName)):
      if(args.printlevel>0):
        print('Saving true image: ',trueFileName)
      save_true(trueFileName,args.outputdir,relative=relativeScale)
  
  # Read Velocity Encoding if file exists
  vencfile = args.maindir+'venc_n1.npy'
  if(os.path.exists(vencfile)):
    venc = np.load(vencfile)
  else:
    venc = None

  # Loop over undersamplingn ratios
  for p in [0.25, 0.5, 0.75, 0.80, 0.85, 0.90, 0.95]:
    for samptype in ['bernoulli', 'vardengauss']:
      if(args.savemask):
        maskFile = args.maskdir + 'undersamplpattern_p' + str(int(p*100)) + samptype + '_n' + str(args.numsamples) + '.npy'
        if(os.path.exists(maskFile)):
          maskoutFile = args.outputdir + prefstr + 'undersamplpattern_p' + str(int(p*100)) + samptype + '_n' + str(args.numsamples) + '.png'
          if(args.printlevel > 0):
            print('Saving undersampling mask: ',maskFile)
          save_mask(maskFile, maskoutFile)           

      for noise_percent in [0.0, 0.01, 0.05, 0.1, 0.3]:
        for wavetype in ['HAAR','DB8']:
          # Loop on the reconstruction method
          for algtype in ['CS','CSDEB','OMP']:
            if(args.saverec):
              recnpy = args.recdir + 'rec_noise' + str(int(noise_percent*100)) + \
                                            '_p' + str(int(p*100)) + samptype + \
                                            '_n' + str(args.numsamples) + \
                                            '_w' + str(wavetype) + \
                                            '_a' + str(algtype) + '.npy'
              if(os.path.exists(recnpy)):
                if(args.printlevel > 0):
                  print('Saving image reconstructions: ',recnpy)
                save_rec(recnpy, venc, p, samptype, noise_percent, wavetype, algtype, prefstr, args.maindir, args.outputdir, args.limits, args.fluidmaskfile, singleChannel=args.singlechannel,relative=relativeScale)
                if(args.printlevel > 0):
                  print('Saving image reconstruction errors')
                save_rec_noise(recnpy, trueFileName, venc, p, samptype, noise_percent, wavetype, algtype, prefstr, args.outputdir, use_truth=args.usetrueasref, singleChannel=args.singlechannel)

          # Save linear reconstructions if appropriate
          if(os.path.exists(recnpy)):
            if(args.savelin):
              if(args.printlevel > 0):
                print('Saving linear reconstruction')
              save_lin(recnpy, venc, p, samptype, noise_percent, wavetype, args.numsamples, prefstr, args.maindir, args.outputdir, args.limits, args.fluidmaskfile, args.usemultipatterns, singleChannel=args.singlechannel,relative=relativeScale)

# MAIN 
if __name__ == '__main__':

  # Init parser
  parser = argparse.ArgumentParser(description='Generate result images.')

  # numRealizations
  parser.add_argument('-n', '--numsamples',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=100,
                      type=int,
                      choices=None,
                      required=False,
                      help='number of repetitions',
                      metavar='',
                      dest='numsamples')

  # maindir
  parser.add_argument('-m', '--maindir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder with original image and velocity encoding file',
                      metavar='',
                      dest='maindir')

  # recdir
  parser.add_argument('-r', '--recdir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder with the reconstructed images',
                      metavar='',
                      dest='recdir')

  # maskdir
  parser.add_argument('-s', '--maskdir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder containing the undesampling masks',
                      metavar='',
                      dest='maskdir')

  # outputdir
  parser.add_argument('-o', '--outputdir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='image output folder',
                      metavar='',
                      dest='outputdir')

  # output image prefix
  parser.add_argument('--imgprefix',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='',
                      type=str,
                      choices=None,
                      required=False,
                      help='output image prefix',
                      metavar='',
                      dest='imgprefix')

  # Use the True Images as a refences when evaluating noise
  parser.add_argument('--usetrueasref',
                      action='store_true',
                      default=False,
                      required=False,
                      help='Use original image instead of average image to evaluate noise',
                      dest='usetrueasref')

  # save_true
  parser.add_argument('--savetrue',
                      action='store_true',
                      default=False,
                      required=False,
                      help='save original image',
                      dest='savetrue')

  # save_mask
  parser.add_argument('--savemask',
                      action='store_true',
                      default=False,
                      required=False,
                      help='save undersampling mask',
                      dest='savemask')

  # save_rec
  parser.add_argument('--saverec',
                      action='store_true',
                      default=False,
                      required=False,
                      help='save reconstructed images',
                      dest='saverec')

  # save_noise
  parser.add_argument('--savenoise',
                      action='store_true',
                      default=False,
                      required=False,
                      help='save noise image',
                      dest='savenoise')

# save linear reconstructions
  parser.add_argument('--savelin',
                      action='store_true',
                      default=False,
                      required=False,
                      help='save linear reconstructions',
                      dest='savelin')

  # singlechannel
  parser.add_argument('--singlechannel',
                      action='store_true',
                      default=False,
                      required=False,
                      help='treat the image as single-channel, without velocity components',
                      dest='singlechannel')

  # Print Level
  parser.add_argument('-p', '--printlevel',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=0,
                      type=float,
                      choices=None,
                      required=False,
                      help='print level, 0 - no print, >0 increasingly more information ',
                      metavar='',
                      dest='printlevel')

  # fluidmaskfile
  parser.add_argument('--fluidmaskfile',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='',
                      type=str,
                      choices=None,
                      required=False,
                      help='name of the npy file containing the binary mask',
                      metavar='',
                      dest='fluidmaskfile')   

  # use unique undesampling pattern for each reconstruction (false) or change every time
  parser.add_argument('-um', '--usemultipatterns',
                      action='store_true',
                      default=False,
                      required=False,
                      help='generate a unique undersampling pattern for each noise realization',
                      dest='usemultipatterns')

  # limits
  parser.add_argument('--limits',
                      action=None,
                      nargs='*',
                      const=None,
                      default=[None,None,None,None,None,None,None,None],
                      type=float,
                      choices=None,
                      required=False,
                      help='list of images limits',
                      metavar='',
                      dest='limits')


  # Parse Commandline Arguments
  args = parser.parse_args()

  # Save all the images
  save_all(args)

  if(args.printlevel > 0):
    print('Completed!!!')  




# %% 
import numpy as np
import numpy.linalg as la
import time
import scipy.optimize as sciopt
from solver_l1_norm import sft, project_l1_ball, MinimizeSumOfSquares

# %%
# Projection onto the nuclear norm ball / tall matrices
def project_nuclear_ball(X, t, eps=1E-5):
    # Compute right singular vectors
    s, V = la.eigh(X.conj().T @ X)
    s = np.sqrt(np.abs(s))

    if np.sum(s) <= t:
        return X

    idx = np.where(s > eps)[0]
    V = V[:, idx]
    s = s[idx]

    st = project_l1_ball(s, t)
    return X @ ((V * (st/s) ) @ V.conj().T)

# Singular value thresholding
def svt(X, t, eps=1E-6):
    # Compute right singular vectors
    s, V = la.eigh(X.conj().T @ X)
    s = np.sqrt(np.abs(s))

    idx = np.where(s >= t)[0]
    if idx.size == 0:
        return np.zeros(shape=X.shape, dtype=X.dtype)

    idx = np.where(s > eps)[0]
    V = V[:, idx]
    s = s[idx]

    st = sft(s, t)
    return X @ ((V * (st/s) ) @ V.conj().T)

# Minimize Sum-of-Squares subject to nuclear norm constraints
#   min_X || A(X) - y ||_2^2 s.t. ||X||_* <= t
def MinimizeSumOfSquaresNuclearBall(t, y, A, Xinit=None, L=None,
                                        maxItns=1E4, 
                                        dxAbsTol=1E-4, dxRelTol=1E-6,
                                        dpAbsTol=1E-5, dpRelTol=1E-8,
                                        disp=False, printEvery=10,
                                        restart=True):
    # Project initial iterate on the nuclear-ball
    if Xinit is None:
        X = np.zeros(A.inShape, dtype=np.complex)
    else:
        X = Xinit
    X = project_nuclear_ball(X, t)
    # Lipschitz constant
    if L is None:
        L = 2.05 * A.norm() ** 2
    # Variables
    rX = A.eval(X) - y
    fX = la.norm(rX.ravel()) ** 2
    gX = 2.0 * A.adjoint(rX)
    rpX = X - project_nuclear_ball(X - gX, t)
    rpXNrm = la.norm(rpX.ravel())
    # Initialize variables
    XNrm = la.norm(X.ravel())
    dXNrm = np.inf
    dfX = np.inf
    itn = 0
    s = 1.0
    Z = X
    # Optimization loop
    stop = False
    if disp:
        print('[PGD-SoS-NuclearBall]')
        print(' Initial objective:   {:1.6e}'.format(fX))
        print(' Initial residual:    {:1.6e}'.format(rpXNrm))
    while not stop:
        itn = itn + 1
        gZ = 2.0 * A.adjoint(A.eval(Z) - y)
        Xp = project_nuclear_ball(Z - gZ/L, t)
        sp = 0.5 * (1.0 + np.sqrt(1.0 + 4.0 * s ** 2))
        Zp = Xp + ((s - 1.0)/sp) * (Xp - X)

        # Update
        rXp = A.eval(Xp) - y
        fXp = la.norm(rXp.ravel()) ** 2

        dfX = fXp - fX
        dXNrm = la.norm((Xp - X).ravel())

        if restart and dfX > 1E-12:
            Z = X
            s = 1.0
            dfX = np.inf
            dXNrm = np.inf
        else:
            X = Xp
            gX = 2.0 * A.adjoint(rXp)
            rpX = X - project_nuclear_ball(X - gX, t)
            rpXNrm = la.norm(rpX.ravel())

            Z = Zp
            fX = fXp
            s = sp

        XNrm = la.norm(X.ravel())
        dXTol = np.max([ dxAbsTol, dxRelTol * XNrm ])
        rpXTol = np.max([ dpAbsTol, dpRelTol * XNrm ]) 
        if (dXNrm < dXTol and rpXNrm < rpXTol) or itn > maxItns:
            stop = True
        if disp and (itn == 1 or np.mod(itn, printEvery) == 0 or stop):
            print(' {:04d} | fX: {:1.3E} | dfX: {:+1.3E} | dXNrm: {:1.3E} | rpXNrm: {:5.3E}'.format(itn, fX, dfX, dXNrm, rpXNrm))
    return X, fX

# Minimize nuclear norm regularized Sum-of-Squares
#   min_X || A(X) - y ||_2^2 + t ||X||_*
def MinimizeNNDN(t, y, A, Xinit=None, L=None,
                    maxItns=1E4, 
                    dxAbsTol=1E-4, dxRelTol=1E-6,
                    dpAbsTol=1E-5, dpRelTol=1E-8,
                    disp=False, printEvery=10,
                    restart=True):
    # Project initial iterate on the nuclear-ball
    if Xinit is None:
        X = np.zeros(A.inShape, dtype=np.complex)
    else:
        X = Xinit
    # Lipschitz constant
    if L is None:
        L = 2.05 * A.norm() ** 2
    # Variables
    rX = A.eval(X) - y
    fX = la.norm(rX.ravel()) ** 2 + t * la.norm(np.sqrt(la.eigvalsh(X.conj().T @ X)), 1)
    gX = 2.0 * A.adjoint(rX)
    rpX = X - svt(X - gX, t)
    rpXNrm = la.norm(rpX.ravel())
    # Initialize variables
    XNrm = la.norm(X.ravel())
    dXNrm = np.inf
    dfX = np.inf
    itn = 0
    s = 1.0
    Z = X
    # Optimization loop
    stop = False
    if disp:
        print('[PGD-NNDN]')
        print(' Initial objective:   {:1.6e}'.format(fX))
        print(' Initial residual:    {:1.6e}'.format(rpXNrm))
    while not stop:
        itn = itn + 1
        gZ = 2.0 * A.adjoint(A.eval(Z) - y)
        Xp = svt(Z - gZ/L, t/L)
        sp = 0.5 * (1.0 + np.sqrt(1.0 + 4.0 * s ** 2))
        Zp = Xp + ((s - 1.0)/sp) * (Xp - X)

        # Update
        rXp = A.eval(Xp) - y
        fXp = la.norm(rXp.ravel()) ** 2 + t * la.norm(np.sqrt(la.eigvalsh(Xp.conj().T @ Xp)), 1)

        dfX = fXp - fX
        dXNrm = la.norm((Xp - X).ravel())
        if restart and dfX > 1E-12:
            Z = X
            s = 1.0
            dfX = -np.inf
            dXNrm = np.inf
        else:
            X = Xp
            gX = 2.0 * A.adjoint(rXp)
            rpX = X - svt(X - gX, t)
            rpXNrm = la.norm(rpX.ravel())

            Z = Zp
            fX = fXp
            s = sp

        dXTol = np.max([ dxAbsTol, dxRelTol * XNrm ])
        rpXTol = np.max([ dpAbsTol, dpRelTol * XNrm ]) 
        if (dXNrm < dXTol and rpXNrm < rpXTol) or itn > maxItns:
            stop = True

        if disp and (itn == 1 or np.mod(itn, printEvery) == 0 or stop):
            print(' {:04d} | obj: {:1.3E} | dfx: {:+1.3E} | dxNrm: {:1.3E} | rpxNrm: {:5.3E}'.format(itn, fX, dfX, dXNrm, rpXNrm))
    return X, fX

class RootSolverNuclearNormNoisy():
    def __init__(self, eta, y, A,
                    maxItns=1E4, 
                    dxAbsTol=1E-4, dxRelTol=1E-6,
                    dpAbsTol=1E-5, dpRelTol=1E-8,
                    disp=False, printEvery=10,
                    restart=True,
                    method='NNDN'):
        # Problem parameters
        self.eta = eta
        self.y = y
        self.A = A

        self.ANrm = A.norm()
        self.L = 2.05 * self.ANrm ** 2

        # For bracketing
        self._itn = 0
        self._t = [ None, None ]
        self._ft = [ None, None ]
        self._Xinit = [ None, None ]
        if method == 'SoS-NucBall':
            # Lower bound
            self._t[0] = 0.0
            self._ft[0] = la.norm(y.ravel()) - eta
            self._Xinit[0] = np.zeros(A.inShape, dtype=np.complex)
            # Upper bound
            _Xopt, _fopt = MinimizeSumOfSquares(y, A, 
                                                maxItns=maxItns, 
                                                dxAbsTol=dxAbsTol, dxRelTol=dxRelTol,
                                                gradAbsTol=dpAbsTol, gradRelTol=dpRelTol,
                                                disp=False)
            self._t[1] = la.norm(np.sqrt(la.eigvalsh(_Xopt.conj().T @ _Xopt)), 1)
            self._ft[1] = la.norm((A.eval(_Xopt) - y).ravel(), 2) - self.eta
            self._Xinit[1] = _Xopt
        elif method == 'NNDN':
            # Lower bound
            _Xopt, _fopt = MinimizeSumOfSquares(y, A, 
                                                maxItns=maxItns, 
                                                dxAbsTol=dxAbsTol, dxRelTol=dxRelTol,
                                                gradAbsTol=dpAbsTol, gradRelTol=dpRelTol,
                                                disp=False)            
            self._t[0] = 0.0
            self._ft[0] = -eta
            self._Xinit[0] = _Xopt
            # Upper bound
            _Z = A.adjoint(y)
            self._t[1] = la.norm(np.sqrt(la.eigvalsh(_Z.conj().T @ _Z)), np.inf)
            self._ft[1] = la.norm(y.ravel(), 2) - self.eta
            self._Xinit[1] = np.zeros(A.inShape, dtype=np.complex)
        else:
            raise ValueError('Method can be either SoS-NucBall or NNDN.')
        self.method = method
        # Solver options
        self.restart = restart
        # Solver parameters
        self.maxItns = maxItns
        self.dxAbsTol = dxAbsTol
        self.dxRelTol = dxRelTol
        self.dpAbsTol = dpAbsTol
        self.dpRelTol = dpRelTol
        # Display options
        self.disp = disp
        self.printEvery = printEvery

    def solve(self, t):
        self._itn = self._itn + 1
        if self.disp:
            print(' {:02d} | t* = {:1.4E} in [{:1.4E}, {:1.4E}] | f: [{:+1.4E}, {:+1.4E}]'.format(self._itn, t, self._t[0], self._t[1], self._ft[0], self._ft[1]))

        if np.abs(t - self._t[0]) < np.abs(t - self._t[1]):
            tNear = 0
        else:
            tNear = 1

        theta = np.abs(t - self._t[0]) / (np.abs(t - self._t[0]) + np.abs(t - self._t[1]))
        Xinit = theta * self._Xinit[0] + (1.0 - theta) * self._Xinit[1]

        if self.method == 'SoS-NucBall':
            Xopt, _ = MinimizeSumOfSquaresNuclearBall(t, y, A, Xinit=Xinit, L=self.L,
                                                        maxItns=self.maxItns, 
                                                        dxAbsTol=self.dxAbsTol, dxRelTol=self.dxRelTol,
                                                        dpAbsTol=self.dpAbsTol, dpRelTol=self.dpRelTol,
                                                        disp=False, printEvery=self.printEvery,
                                                        restart=self.restart)
        elif self.method == 'NNDN':
            Xopt, _ = MinimizeNNDN(t, y, A, Xinit=Xinit, L=self.L,
                                    maxItns=self.maxItns, 
                                    dxAbsTol=self.dxAbsTol, dxRelTol=self.dxRelTol,
                                    dpAbsTol=self.dpAbsTol, dpRelTol=self.dpRelTol,
                                    disp=False, printEvery=self.printEvery,
                                    restart=self.restart)
        
        if t > self._t[1]:
            self._t[0] = self._t[1]
            self._ft[0] = self._ft[1]
            self._Xinit[0] = self._Xinit[1]

            self._t[1] = t
            self._ft[1] = la.norm((A.eval(Xopt) - y).ravel(), 2) - self.eta
            self._Xinit[1] = Xopt
        elif t < self._t[0]:
            self._t[1] = self._t[0]
            self._ft[1] = self._ft[0]
            self._Xinit[1] = self._Xinit[0]

            self._t[0] = t
            self._ft[0] = la.norm((A.eval(Xopt) - y).ravel(), 2) - self.eta
            self._Xinit[0] = Xopt
        else:
            self._t[tNear] = t
            self._ft[tNear] = la.norm((A.eval(Xopt) - y).ravel(), 2) - self.eta
            self._Xinit[tNear] = Xopt

        return la.norm((A.eval(Xopt) - y).ravel(), 2) - self.eta

def RecoveryNuclearNormNoisy(eta, y, A, 
                    maxItns=1E4, 
                    dxAbsTol=1E-4, dxRelTol=1E-6,
                    dpAbsTol=1E-5, dpRelTol=1E-8,
                    disp=False, printEvery=100,
                    method='NNDN'):

    # Create solver object
    T = RootSolverNuclearNormNoisy(eta, y, A,
                                    maxItns=maxItns, 
                                    dxAbsTol=dxAbsTol, dxRelTol=dxRelTol,
                                    dpAbsTol=dpAbsTol, dpRelTol=dpRelTol,
                                    disp=disp, printEvery=printEvery,
                                    restart=True, method=method)
    if disp:
        print('[RECOVERY: NUCLEAR NORM-NOISY]' )
        print(' Method:          {:s}'.format(method) )
        print(' Bound:           {:1.4E}'.format(eta) )
        print(' Initial values:')
        print('     t           [{:1.4E},   {:1.4E}]'.format(T._t[0], T._t[1]))
        print('     f(t)        [{:+1.4E}, {:+1.4E}]'.format(T._ft[0], T._ft[1]))
    # Root finding via TOMS748
    tStart = time.time()
    tMin = T._t[0]
    tMax = T._t[1]
    topt, rnfo = sciopt.toms748(lambda t: T.solve(t), tMin, tMax, xtol=5E-4, full_output=disp, disp=disp)
    if np.abs(topt - T._t[0]) < np.abs(topt - T._t[1]):
        Xopt = T._Xinit[0]
    else:
        Xopt = T._Xinit[1]
    tStop = time.time() - tStart
    if disp:
        print('  Summary');
        print('      Iterations:          {:d}'.format(rnfo.iterations) )
        print('      Function calls:      {:d}'.format(rnfo.function_calls) )
        print('      Optimal value:       {:5.3E}'.format(topt) )
        print('      Elapsed time:        {:8.4f} seconds'.format(tStop) )
    return Xopt, la.norm(Xopt.ravel(), 1)

# %%
if __name__ == '__main__':
    # Import
    from maps import OperatorFourierLowRank
    
    print('[TESTING SOLVER ROUTINES]')
    # *** EXPERIMENTS FOR SINGLE IMAGE AND SUBSAMPLING - L1-NORM
    # Set up map
    imShape = (128, 128)
    samplingSet = np.where(np.random.uniform(size=imShape) < 0.75, True, False)
    A = OperatorFourierLowRank(imShape, samplingSet=samplingSet)
    xinit = np.random.normal(size=A.inShape) + 1j * np.random.normal(size=A.inShape)
    y = A.eval(xinit)
    print('[SINGLE IMAGE - RANDOM UNDERSAMPLING - L1-NORM RECOVERY]')
    print('    Input shape:    {:d} x {:d}'.format(A.inShape[0], A.inShape[1]))
    print('    Output shape:   {:d} x {:d}'.format(A.outShape[0], 1))
    print('    Operator norm:  {:1.3e}'.format(A.norm()))
    # Tests
    print('\n[TEST: MinimizeSumOfSquares]')
    xsol, fmin = MinimizeSumOfSquares(y, A, disp=True)
    print('\n[TEST: MinimizeSumOfSquaresNuclearBall]')
    print('  ---> with no restart...')
    xsol, fmin = MinimizeSumOfSquaresNuclearBall(1.0, y, A, disp=True, restart=False, printEvery=10)
    print('  ---> with restart...')
    _xsol, _fmin = MinimizeSumOfSquaresNuclearBall(1.0, y, A, disp=True, restart=True)
    print('  ---> Difference:     {:1.5e}'.format(la.norm((xsol - _xsol).ravel())))
    print('\n[TEST: RecoveryL1Noisy - SoS-NucBall]')
    xsol, fmin = RecoveryNuclearNormNoisy(2.0, y, A, maxItns=1E4, disp=True, printEvery=100, method='SoS-NucBall')
    print('\n[TEST: RecoveryL1Noisy - NNDN]')
    _xsol, _fmin = RecoveryNuclearNormNoisy(2.0, y, A, maxItns=1E4, disp=True, printEvery=100, method='NNDN')
    print('  ---> Difference:')
    print('         Objective:    {:1.5e}'.format(np.abs(fmin - _fmin)))
    print('         Solution:     {:1.5e}'.format(la.norm((xsol - _xsol).ravel())))

# %%



import sys,os
import random
import numpy as np
import matplotlib.pyplot as plt
sys.path.append('../')
from CSRecoverySuite import extractFluidMask
from recover import recover_vel
import argparse

home = os.getenv('HOME')

# Parameters
EPSILON = 0.5

def circle(shape, center, dist, useFluidMask=False, fluidMask=None):
  # probably a more efficient way to determine this
  pts = []
  for x in range(max(0,center[0]-dist), min(center[0]+dist, shape[0])):
    for y in range(max(0,center[1]-dist), min(center[1]+dist, shape[1])):
      if abs((x-center[0])**2 + (y-center[1])**2 - dist**2) < EPSILON**2:
        if(useFluidMask):
          if(fluidMask[x,y]):
            pts.append([x, y])
        else:
          pts.append([x, y])
  return pts

def pick_random_point(imsz, useFluidMask=False, fluidMask=None):
  if(useFluidMask):
    found = False
    while(not found):      
      pt1 = (random.randint(0, imsz[0]-1), random.randint(0, imsz[1]-1))
      found = fluidMask[pt1[0],pt1[1]]
  else:
    pt1 = (random.randint(0, imsz[0]-1), random.randint(0, imsz[1]-1))
  return pt1

def select_points(dist, imsz, useFluidMask=False, fluidMask=None):
  # Select center at random
  pt1 = pick_random_point(imsz, useFluidMask, fluidMask)
  choices = circle(imsz, pt1, dist, useFluidMask, fluidMask)
  # Not sure why this loop is here...
  while not choices: # make sure there were points within range for distance
    pt1 = pick_random_point(imsz, useFluidMask, fluidMask)
    choices = circle(imsz, pt1, dist, useFluidMask, fluidMask)
  # Select pt2 at random from the list
  pt2 = tuple(random.choice(choices))
  if pt1 <= pt2: # pt1[0] < pt2[0] or (pt1[0] == pt2[0] and pt1[1] <= pt2[1]):
    return pt1,pt2
  return pt2, pt1

def get_points(size, num_pts, imsz, ptsdir, pointFileName, useFluidMask=False, fluidMask=None):
  print('Getting points of size ' + str(size) + ' and ' + str(num_pts) + ' points...')
  points = np.zeros((size, num_pts, 2, 2), dtype=int)
  for k in range(1,size+1):
    if(k % 10 == 0):
      print(k)
    for j in range(num_pts):
      pt1, pt2 = select_points(k,imsz,useFluidMask,fluidMask)
      while pt1[0] in points[k-1,:,0,0] and pt1[1] in points[k-1,:,0,1] and pt2[0] in points[k-1,:,1,0] and pt2[0] in points[k-1,:,1,1]:
        # Select new points if we've already selected these
        pt1, pt2 = select_points(k,imsz,useFluidMask,fluidMask) 
      points[k-1, j, 0, 0] = pt1[0]
      points[k-1, j, 0, 1] = pt1[1]
      points[k-1, j, 1, 0] = pt2[0]
      points[k-1, j, 1, 1] = pt2[1]
  # Save Point File
  np.save(pointFileName,points)

def get_samples(recfile, singlechannel, vencdir):
  # Load reconstructed images
  samples = np.load(recfile)
  # Get Vencoding
  vencfile = vencdir + 'venc_n1' + '.npy'
  if os.path.exists(vencfile):
    venc = np.load(vencfile)
  else:
    venc = None
  # Recover the velocities
  if(singlechannel):
    samples = np.absolute(samples)
  else:
    samples = recover_vel(samples, venc)
  # Compute average reconstructed image
  avgimg = np.mean(samples, axis=0)
  # Compute reconstruction noise and return
  for k in range(samples.shape[0]):
    samples[k] = samples[k] - avgimg    
  return samples

def get_saved_points(samples, size, num_pts, maindir, ptsdir, useFluidMask, fluidMaskFile):
  imsz = samples.shape[3:]

  # Get File Name
  fluidMask = None
  if(useFluidMask):
    if(fluidMaskFile == ''):
      print('Using fluid mask...')
    else:
      print('Using fluid mask from file: '+maindir+fluidMaskFile+'...')

    # Get File with Original Image
    orig_file = maindir + 'imgs_n1.npy'
  
    # Check if files exist and open the images
    if(os.path.exists(orig_file)):
      orig = np.load(orig_file).astype(np.complex)
    else:
      print('ERROR: Cannot find file with original image.')
      exit(-1)

    if(fluidMaskFile == ''):
      # Extract fluid mask for simple geometries
      fluidMask = extractFluidMask(orig)
    else:
      # Read mask from file 
      fluidMask = np.load(maindir + fluidMaskFile)

    if(False):
      # Plot the fluid mask for debug
      plt.imshow(fluidMask,cmap='gray')
      plt.show()
      exit()

    # Set File Name for Point in the Fluid Mask
    pointFileName = ptsdir+'points_s' + str(size) + '_n' + str(num_pts) + '_fluidmask.npy'
  else:

    # Set the file name for points without fluid mask
    pointFileName = ptsdir+'points_s' + str(size) + '_n' + str(num_pts) + '.npy'

  # Read or generate the points
  if not os.path.isfile(pointFileName):
    get_points(size, num_pts, imsz, ptsdir, pointFileName, useFluidMask, fluidMask)
  points = np.load(pointFileName)
  return points

def get_coeff(size, num_pts, samples, points):
  print('in coeff', samples.shape)
  coeff = np.zeros((samples.shape[1], size, num_pts))
  for k in range(1, size+1):
    for j in range(num_pts):
      pt1 = (points[k-1, j, 0,0], points[k-1,j, 0,1])
      pt2 = (points[k-1, j, 1,0], points[k-1,j, 1,1])
      for v in range(samples.shape[1]):
        var1 = np.abs(samples[:, v, 0, pt1[0], pt1[1]])
        var2 = np.abs(samples[:, v, 0, pt2[0], pt2[1]])
        # If both sequences contain zeros, this generates a nan.
        # We replace nan with ones. This is only a problem when evaluating the 
        # correlation between pixels in the background. 
        coeff[v, k-1, j] = np.corrcoef(np.asarray([var1, var2]))[0,1] # correlation
  if(np.isnan(coeff).any()):
    print('WARNING: Found some NAN, probably due to sequences of zeros in the background. Replacing with correlation of 1.0')
    coeff[np.isnan(coeff)] = 1.0
  return coeff

def get_vals(recfile, savefile, maxcorrpixeldist, numpts, singlechannel, maindir, recdir, ptsdir, useFluidMask, fluidMaskFile, save_numpy=True):
  samples = get_samples(recfile, singlechannel, args.vencdir)
  print('Use Fluid Mask: ',useFluidMask)
  points  = get_saved_points(samples, maxcorrpixeldist, numpts, maindir, ptsdir, useFluidMask, fluidMaskFile)
  coeff   = get_coeff(maxcorrpixeldist, numpts, samples, points)
  if save_numpy:
    np.save(savefile, coeff)
  return coeff

def get_all(args):
  # size is max. distance to retrieve correlations for
  # num_pts is the number of points to average correlation across
  for noise_percent in [0.0, 0.01, 0.05, 0.1, 0.3]:
    for p in [0.25, 0.5, 0.75, 0.80, 0.85, 0.90, 0.95]:
      for wavetype in ['HAAR','DB8']:
        for algtype in ['CS','CSDEB','OMP']:
          for samptype in ['bernoulli', 'vardengauss']:
            recfile = args.recdir + 'rec_noise' + str(int(noise_percent*100)) + \
                                    '_p' + str(int(p*100)) + samptype + \
                                    '_n' + str(args.numsamples) + \
                                    '_w' + str(wavetype) + \
                                    '_a' + str(algtype) + '.npy'

            savefile = args.recdir + 'corrcoeff' + str(args.numpts) + \
                                     '_noise' + str(int(noise_percent*100)) + \
                                     '_p' + str(int(p*100)) + samptype + \
                                     '_n'+str(args.numsamples)+ \
                                     '_w' + str(wavetype) + \
                                     '_a' + str(algtype)
            
            # Change name for fluid Mask 
            if(args.usefluidmask):
              savefile += '_fluid.npy'
            else:
              savefile += '.npy'

            if(os.path.exists(recfile)):
              get_vals(recfile, savefile, args.maxcorrpixeldist, args.numpts, args.singlechannel, args.maindir, args.recdir, args.ptsdir, args.usefluidmask, args.fluidmaskfile)
              print('Saved! Noise: ', noise_percent,', undersampling prob: ', p, ', undersampling mask: ', samptype, ', wavetype: ', wavetype, ', algorithm: ', algtype)

# MAIN 
if __name__ == '__main__':

  # Init parser
  parser = argparse.ArgumentParser(description='Generate result images.')

  # numsamples
  parser.add_argument('-n', '--numsamples',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=100,
                      type=int,
                      choices=None,
                      required=False,
                      help='number of repetitions',
                      metavar='',
                      dest='numsamples')
  # numpts
  parser.add_argument('-m', '--numpts',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=50,
                      type=int,
                      choices=None,
                      required=False,
                      help='number of point pairs for computing correlations',
                      metavar='',
                      dest='numpts')
  # size
  parser.add_argument('-c', '--maxcorrpixeldist',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=100,
                      type=int,
                      choices=None,
                      required=False,
                      help='Maximum integer pixel distance to use for correlation computation',
                      metavar='',
                      dest='maxcorrpixeldist')
  # recdir
  parser.add_argument('-d', '--maindir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder containing the original image',
                      metavar='',
                      dest='maindir')  
  # recdir
  parser.add_argument('-r', '--recdir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder containing the non linear reconstructions',
                      metavar='',
                      dest='recdir')
  # ptsdir
  parser.add_argument('-t', '--ptsdir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder containing the saved point coordinates',
                      metavar='',
                      dest='ptsdir')
  # vencdir
  parser.add_argument('-v', '--vencdir',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='./',
                      type=str,
                      choices=None,
                      required=False,
                      help='folder containing the velocity encoding',
                      metavar='',
                      dest='vencdir')
  # singlechannel
  parser.add_argument('--singlechannel',
                      action='store_true',
                      default=False,
                      required=False,
                      help='treat the image as single-channel, without velocity components',
                      dest='singlechannel')  
  # usefluidmask
  parser.add_argument('--usefluidmask',
                      action='store_true',
                      default=False,
                      required=False,
                      help='compute correlation only within fluid region',
                      dest='usefluidmask')  

  # fluidmaskfile
  parser.add_argument('--fluidmaskfile',
                      action=None,
                      # nargs='+',
                      const=None,
                      default='',
                      type=str,
                      choices=None,
                      required=False,
                      help='name of the npy file containing the binary mask',
                      metavar='',
                      dest='fluidmaskfile')  

  # Print Level
  parser.add_argument('-p', '--printlevel',
                      action=None,
                      # nargs='+',
                      const=None,
                      default=0,
                      type=float,
                      choices=None,
                      required=False,
                      help='print level, 0 - no print, >0 increasingly more information ',
                      metavar='',
                      dest='printlevel')

  # Parse Commandline Arguments
  args = parser.parse_args()

  # Save all the images
  get_all(args)

  # Completed!
  if(args.printlevel > 0):
    print('Completed!!!')  



import numpy.fft as fft
import pywt
import numpy as np
import sys
import numpy.linalg as la
from scipy import optimize
from scipy.optimize import toms748
from scipy.stats import bernoulli
import scipy
import math
import time

def get_umask_string(samptype):
  if(samptype == 'bernoulli'):
    return 'Bernoulli'
  elif(samptype == 'vardengauss'):
    return 'Gauss'
  else:
    print('ERROR: Invalid mask type')
    sys.exit(-1)

def extractFluidMask(img):  
  res = np.max(np.absolute(img),axis=0)[0,0]
  return (res > 0)    

def get_method_string(method):
  if(method == 'cs'):
    return 'CS'
  elif(method == 'csdebias'):
    return 'CSDEB'
  elif(method == 'omp'):
    return 'OMP'    
  else:
    print('ERROR: Invalid recovery method')
    sys.exit(-1)

def get_wavelet_string(wavelet):
  if(wavelet == 'haar'):
    return 'HAAR'
  elif(wavelet == 'db8'):
    return 'DB8'
  else:
    print('ERROR: Invalid wavelet type')
    sys.exit(-1)

def toSlice(idx):
  if idx == None:
    return slice(None)
  else:
    return idx

# Methods for cropping in the case the dimensions aren't a power of 2
def powerof2(num):
  # Return the highest power of 2 less than or equal to number
  return int(math.pow(2,math.floor(math.log(num, 2))))

def crop(x):
  if x.shape[1] > x.shape[0]:
    dim1 = powerof2(x.shape[0])
    #max between power of 2 or multiple of other dimension
    dim2 = max(powerof2(x.shape[1]), math.floor(x.shape[1]/dim1)*dim1)
  else:
    dim2 = powerof2(x.shape[1])
    dim1 = max(powerof2(x.shape[0]), math.floor(x.shape[0]/dim2)*dim2)
  return x[0:dim1, 0:dim2]

def isvalidcrop(x):
  '''
  Check if the image has valid power of 2 dimensions
  '''
  y = x.copy()
  y = crop(y)
  if(x.shape[0] == y.shape[0])and(x.shape[1] == y.shape[1]):
    return True
  else:
    return False

def pywt2array( x, shape ):
  # Essentially a breadth-first traversal
  y = np.zeros((0,1))
  for sublist in x:
    for z in sublist:
      y = np.concatenate((y, np.expand_dims(z.ravel(), axis=1)),axis=0)
  y = y.reshape(shape)
  return y

def array2pywt( x ):
  shape = x.shape
  # Assuming both dims power of 2 or one dim is multiple of other (which is power of 2)
  maxlevel = int(math.log(min(shape),2))
  x = x.ravel()
  coeffs = [None]*(maxlevel+1)
  i = len(x)
  if shape[0] > shape[1]:
    initdim1 = int(shape[0]/shape[1])
    initdim2 = 1
  else:
    initdim1 = 1
    initdim2 = int(shape[1]/shape[0])
  coeffs[0] = np.reshape(x[0:initdim1*initdim2],(initdim1,initdim2))
  for k in reversed(range(1,maxlevel+1)):
    dim1 = int(initdim1*math.pow(2, k-1))
    dim2 = int(initdim2*math.pow(2, k-1))
    coeffs[k] = (np.reshape(x[(i-dim1*dim2*3):(i-dim1*dim2*2)], (dim1,dim2)),
                 np.reshape(x[(i-dim1*dim2*2):(i-dim1*dim2)], (dim1,dim2)),
                 np.reshape(x[(i-dim1*dim2):i], (dim1,dim2)))
    i = i - dim1*dim2*3
  return coeffs

# Added the soft-thresholding function
def sft(x,t):
  # Soft-thresholding
  return x * ( 1 - np.minimum( t / np.maximum( np.abs(x), 1E-32 ), 1 ) )

# Changed the name to project_l1_ball for clarity
def project_l1_ball( x, t ):
  xL1Nrm  = la.norm( x.ravel(), 1 );
  if xL1Nrm > t:
    # If the input is outside the l1-ball we project.
    nx  = x.size;
    # Sort magnitudes in decreasing order
    mx  = np.flip( np.sort( np.absolute( x.ravel() ), kind='mergesort' ) );
    # Obtain unique values (this is needed for some borderline cases)
    umx = np.flip( np.unique( mx ) );
    if( t <= umx[0] - umx[1] ):
      # If the radius of the l1-ball is small and the input is far away
      # and the shrinkage is large.
      vp      = umx[0];
      sroot   = vp - t;
      return sft( x, sroot )
    else:
      # This is the average case
      cmx = np.cumsum( mx );
      smx = cmx - np.array( range(1, nx+1) ) * mx;
      if( smx[nx - 1] < t ):
        # This condition handles some cases when the input is close to
        # the boundary and the shrinkage is small.
        sroot   = (xL1Nrm - t) / nx;
        return sft( x, sroot )
      idp,    = np.where( smx > t );
      idp     = idp[0];
      sroot   = (cmx[idp - 1] - t) / idp;
      return sft( x, sroot )
  else:
    # If the input is within the l1-ball we do nothing.
    return x

class genericOperator(object):
  pass

# Defines a class for linear transforms
class Operator4dFlow(genericOperator):
  
  def __init__(self, insz=None, imsz=None, samplingSet=None, basisSet=None, isTransposed=False, waveletName='haar', waveletMode='periodization'):
    # insz is the size of the array used as input
    self.insz           = insz
    self.imsz           = imsz
    self.waveletName    = waveletName
    self.waveletMode    = waveletMode
    self.isTransposed   = isTransposed
    # SamplingSet is the set of indices that are measured after the mapping
    # has been applied
    self.samplingSet    = samplingSet
    if( samplingSet is None ):
      self.outsz          = imsz
      self._buffer        = None
    else:
      self.outsz          = ( np.count_nonzero( samplingSet ), )
      self._buffer        = np.zeros( self.imsz ) + 1j * np.zeros( self.imsz )
    # Restrict the set of wavelet basis
    self.basisSet = basisSet
    # Constant
    # self._cst           = math.pow( np.prod( insz ), -1/2 );

  def eval(self, x, mode):
    '''
    The method eval implements
      obj.eval(x, 1) returns A * x
      obj.eval(x, 2) returns A' * x where A' is the adjoint
    '''
    if( mode == 1 ): # FORWARD MAP
      if( self.samplingSet is None ):
        # return self._cst * fft.fft2(pywt.waverec2(array2pywt(x), wavelet=self.waveletName, mode=self.waveletMode));
        return fft.fft2(pywt.waverec2(array2pywt(x), wavelet=self.waveletName, mode=self.waveletMode),norm='ortho');
      else:
        # y = self._cst * fft.fft2(pywt.waverec2(array2pywt(x), wavelet=self.waveletName, mode=self.waveletMode));
        y = fft.fft2(pywt.waverec2(array2pywt(x), wavelet=self.waveletName, mode=self.waveletMode),norm='ortho');
        return y[ self.samplingSet ];
    if( mode == 2 ): # ADJOINT MAP
      if( self.samplingSet is None ):
        # arr = np.conj( fft.fft2( np.conj(x) ) )
        # return self._cst * pywt2array(pywt.wavedec2(arr, wavelet=self.waveletName, mode=self.waveletMode), arr.shape);
        arr = np.conj( fft.fft2( np.conj(x),norm='ortho' ) )
        return pywt2array(pywt.wavedec2(arr, wavelet=self.waveletName, mode=self.waveletMode), arr.shape);
      else:
        #y = np.zeros( self.imsz ) + 1j * np.zeros( self.imsz );
        self._buffer[ self.samplingSet ] = x[ : ];
        # arr = np.conj( fft.fft2( np.conj(self._buffer) ) )
        # return self._cst * pywt2array(pywt.wavedec2(arr, wavelet=self.waveletName, mode=self.waveletMode),arr.shape);
        arr = np.conj( fft.fft2( np.conj(self._buffer),norm='ortho' ) )
        return pywt2array(pywt.wavedec2(arr, wavelet=self.waveletName, mode=self.waveletMode),arr.shape);


  #  The method returns the shape of the input array
  def input_size(self):
    if(self.samplingSet is None):
      return self.insz
    else:
      return self.samplingSet.shape

  @property
  def shape(self):
    "The shape of the operator."
    if(self.samplingSet is None):
      shapeX = np.prod(self.outsz)
    else:
      # CHECK !!! it doesn't seem right
      shapeX = np.prod(self.samplingSet.shape)

    if(self.basisSet is None):
      shapeY = np.prod(self.insz)
    else:
      shapeY = len(self.basisSet)

    if(self.isTransposed == False):
      return shapeX,shapeY
    else:
      return shapeY,shapeX

  def __mul__(self, x):
    "Multiplication"
    # Convert x from 1D to 2D
    if(not(self.isTransposed)):
            
      # Create zero vector
      inV = np.zeros(np.prod(self.insz),dtype=np.complex)
      
      # Restrict the set of basis on the input vector x
      if(not( self.basisSet is None)):
        inV[self.basisSet] = x[:]
      else:
        inV[:] = x[:]

      # Create a 2D Wavelet Representation
      inV = inV.reshape(self.insz)

      # Compute the vector y
      # y = self._cst * fft.fft2(pywt.waverec2(array2pywt(inV), wavelet=self.waveletName, mode=self.waveletMode))
      y = fft.fft2(pywt.waverec2(array2pywt(inV), wavelet=self.waveletName, mode=self.waveletMode), norm='ortho')

      # Select frequencies as per sampling set
      if( self.samplingSet is None ):
        return y
      else:
        return y[ self.samplingSet ]
    
    else:
    
      # Apply Fourier Transform Only for frequencies in the sampling set
      if( self.samplingSet is None ):
        # arr = np.conj( fft.fft2( np.conj(x) ) )
        arr = np.conj( fft.fft2( np.conj(x), norm='ortho' ) )
      else:
        self._buffer[ self.samplingSet ] = x[ : ];
        # arr = np.conj( fft.fft2( np.conj(self._buffer) ) )
        arr = np.conj( fft.fft2( np.conj(self._buffer), norm='ortho' ) )

      # Perform wavelet transform
      # res = self._cst * pywt2array(pywt.wavedec2(arr, wavelet=self.waveletName, mode=self.waveletMode),arr.shape)
      res = pywt2array(pywt.wavedec2(arr, wavelet=self.waveletName, mode=self.waveletMode),arr.shape)

      # Filter wavelet coefficients as per basisSet
      if( self.basisSet is None ):
        return res.ravel()
      else:
        return res.ravel()[self.basisSet]

  @property
  def T(self):
    "Transposed of the operator."
    return Operator4dFlow(insz=self.insz, imsz=self.imsz, samplingSet=self.samplingSet, basisSet=self.basisSet, isTransposed=not(self.isTransposed), waveletName=self.waveletName, waveletMode=self.waveletMode)

  # Computes leading singular value of the linear mapping A via power iteration
  def getNorm(self, maxItns=1E3, dsAbsTol=1E-6 ):
    # Initialize variables
    b   = np.random.normal( size = self.input_size() )
    b   = b / la.norm(b)
    s   = 0
    ds  = np.inf
    itn = 0
    # Power iteration loop
    while( ds > dsAbsTol and itn <= maxItns ):
      # Evaluate xk = A'A x
      bk = self.eval(self.eval(b, 1), 2)
      # Evaluate xk' A'Ax to estimate sigma_max^2
      sk = np.sqrt( np.real( np.vdot( bk.flatten(), b.flatten() ) ) )
      ds = np.abs( sk - s )
      # Normalize singular vector
      b  = bk / la.norm(bk)
      s  = sk
    return s

  # Create An operator with a 
  def colRestrict(self,idxSet=None):
    "Restrict operator column set."
    return Operator4dFlow(insz=self.insz, imsz=self.imsz, samplingSet=self.samplingSet, basisSet=idxSet, isTransposed=self.isTransposed, waveletName=self.waveletName, waveletMode=self.waveletMode)

# Computes leading singular value of the linear mapping A via power iteration
def OperatorNorm( A, maxItns=1E3, dsAbsTol=1E-6 ):
  # Initialize variables
  b   = np.random.normal( size = A.input_size() )
  b   = b / la.norm(b)
  s   = 0
  ds  = np.inf
  itn = 0
  # Power iteration loop
  while( ds > dsAbsTol and itn <= maxItns ):
    # Evaluate xk = A'A x
    bk = A.eval(A.eval(b, 1), 2)
    # Evaluate xk' A'Ax to estimate sigma_max^2
    sk = np.sqrt( np.real( np.vdot( bk.flatten(), b.flatten() ) ) )
    ds = np.abs( sk - s )
    # Normalize singular vector
    b  = bk / la.norm(bk)
    s  = sk
  return s

def OperatorTestAdjoint(A, ntrials=100):
  insz    = A.input_size()
  x       = np.random.normal(size=insz) + 1j * np.random.normal(size=insz)
  Ax      = A.eval(x, 1)
  outsz   = Ax.shape
  err     = np.zeros((ntrials,))
  for It in range(0, ntrials):
    x       = np.random.normal(size=insz) + 1j * np.random.normal(size=insz)
    y       = np.random.normal(size=outsz) + 1j * np.random.normal(size=outsz)
    Ax      = A.eval(x, 1)
    Aty     = A.eval(y, 2)
    yAx     = np.sum(np.conjugate(y.ravel()) * Ax.ravel())
    Atyx    = np.sum(np.conjugate(Aty.ravel()) * x.ravel())
    err[It] = np.absolute( yAx - Atyx )
  print('--- ADJOINT TEST')
  print('Inner product')
  print(' Avg. Error:', np.mean(err))
  print(' St. Dev.:', np.std(err))
  print('----------------')

def LSSolver( y, A, w0, maxItns=1E4, dwAbsTol=1E-5, dfwAbsTol=1E-6 ):
    # Find lipschitz constant
    smax        = OperatorNorm( A )
    L           = 1.01 * math.pow(smax, 2)
    # Initialize variables
    w           = w0
    dwNrm       = np.inf
    itn         = 0
    fw          = 0.5 * math.pow(la.norm((y - A.eval(w, 1)).ravel(), 2), 2)
    # Optimization loop
    while( ( dwNrm > dwAbsTol or np.abs(dfw) > dfwAbsTol ) and itn <= maxItns):
        itn         = itn + 1
        wk          = w - (1/L) * A.eval(A.eval(w, 1) - y, 2)
        fwk         = 0.5 * math.pow(la.norm((y - A.eval(wk, 1)).ravel(), 2), 2)
        dfw         = fwk - fw
        dwNrm       = la.norm( (wk - w).ravel(), 2 )
        w           = wk
        fw          = fwk
    return w, fw

def pgdl1( t, x0, L, f, df, maxItns=1E4, dwAbsTol=1E-4, dfwAbsTol=1E-5, disp=0, printEvery=10 ):
    # Project initial iterate on the l1-ball
    w           = project_l1_ball( x0, t );
    fw          = f( w );
    # Initialize variables
    dwNrm       = np.inf;
    dfw         = np.inf;
    itn         = 0;
    s           = 1;
    v           = w;
    # Optimization loop
    while( ( dwNrm > dwAbsTol or np.abs( dfw ) > dfwAbsTol ) and itn < maxItns):
        itn         = itn + 1;
        if( disp == 3 and itn == 1 ):
            print(' [PGDL1] Initial objective: {:+5.3E}'.format(fw) );
        wk          = project_l1_ball( v - (1/L) * df( v ), t );
        sp          = 0.5 * ( 1 + math.pow(1 + 4 * math.pow( s, 2 ), 1/2) );
        vk          = wk + ( ( s - 1 )/sp ) * ( wk - w );
        fwk         = f( wk );
        pwk         = project_l1_ball( wk - (1/L) * df( wk ), t );
        dwNrm       = la.norm( (pwk - wk).ravel(), 2 );
        dfw         = fwk - fw;
        w           = wk;
        v           = vk;
        fw          = fwk;
        s           = sp;
        if( disp == 3 and ( itn == 1 or np.mod( itn, printEvery ) == 0 ) ):
            print(' [PGDL1] Iteration: {:03d} | Objective: {:+5.3E} | Last step: {:+5.3E} | Last decrease: {:+5.3E}'.format(itn, fw, dwNrm, dfw) );
    if( disp > 0 ):
        print(' [PGDL1 Summary] Last iteration: {:03d} | Optimal value: {:+5.3E} | Last step: {:+5.3E} | Last decrease: {:+5.3E}'.format(itn, fw, dwNrm, dfw) );
    return w, fw

class RootPGDL1():
    def __init__(self, eta=0, f=None, df=None, L=None, tmin=None, xmin=None, tmax=None, xmax=None, maxItns=1E4, dwAbsTol=1E-4, dfwAbsTol=1E-5, disp=0):
        self.eta = eta
        self.funObject = f
        self.gradObject = df
        self.lipschitzConstant = L
        self.tmin = tmin
        self.xmin = xmin
        self.tmax = tmax
        self.xmax = xmax
        self.maxItns = maxItns
        self.dwAbsTol = dwAbsTol
        self.dfwAbsTol = dfwAbsTol
        self.disp = disp
        self.lastOptimalPoint = None
        self.lastOptimalValue = None

    def eval(self, t):
        theta = (t - self.tmin) / (self.tmax - self.tmin)
        if( np.absolute( t - self.tmin ) < np.absolute( t - self.tmax ) ):
            xinit = self.xmin
        else:
            xinit = self.xmax
        w, fw = pgdl1(t, x0=xinit, f=self.funObject, df=self.gradObject, L=self.lipschitzConstant, maxItns=self.maxItns, dwAbsTol=self.dwAbsTol, dfwAbsTol=self.dfwAbsTol, disp=0, printEvery=1E32)
        fw = np.sqrt(2 * fw)
        self.lastOptimalPoint = w
        self.lastOptimalValue = fw
        if( np.absolute( t - self.tmin ) < np.absolute( t - self.tmax ) ):
            self.tmin = t
            self.xmin = w
        else:
            self.tmax = t
            self.xmax = w
        if( self.disp > 0 ):
            print('      Trial value t = {:7.5E} with f(t) = {:+7.5E} (theta = {:7.5E})'.format(t, fw - self.eta, theta) );
        return fw - self.eta

def f( y, A, alpha ):
  return 0.5 * math.pow(la.norm((y - A.eval(alpha, 1)).ravel(), 2), 2)

def f_grad( y, A, alpha ):
  return A.eval(A.eval(alpha, 1) - y, 2)

###############################
#### COMPRESSED SENSING #######
###############################

def CSRecovery(eta, y, A, x0, disp=0, printEvery=0):
    # Parameters
    if( disp == 0 ):
        root_disp = 0
    else:
        root_disp = 1
    yNrm        = la.norm( y.ravel(), 2 )
    if( disp > 0 ):
        print(' [CS RECOVERY]' )
        print('      Error level:        {:5.3E}'.format(eta) )
        print('      l2-norm of y:       {:5.3E}'.format(yNrm) )
    if( yNrm <= eta ):
        return np.zeros( x0.shape ), 0
        print('  Summary')
        print('      Optimal value:      {:5.3E}'.format(yNrm) )
        print('      Elapsed time:       {:8.4f} seconds'.format(0.0) )
    # Find lipschitz constant
    smax        = OperatorNorm( A )
    L           = 1.05 * math.pow(smax, 2)
    # Least-squares solution
    wtls, rmax  = LSSolver( y, A, x0 )
    # Lower bound variables
    tmin        = 0
    ftmin       = yNrm
    wtmin       = np.zeros( x0.shape )
    # Find upper bound variables
    wtmax       = wtls
    tmax        = la.norm( wtmax.ravel(), 1 )
    ftmax       = 0
    # Initialize variables
    dt          = tmax - tmin
    dft         = np.abs( ftmin - ftmax )
    if( disp > 0 ):
        print('      Initial tmin:       {:5.3E}'.format(tmin) )
        print('      Initial tmax:       {:5.3E}'.format(tmax) )
    # Root finding via TOMS748
    tend        = time.time()
    froot = RootPGDL1(eta=eta, f=lambda x: f(y, A, x), df=lambda x: f_grad(y, A, x), L=L, tmin=tmin, xmin=wtmin, tmax=tmax, xmax=wtmax, maxItns=1E4, dwAbsTol=1E-4, dfwAbsTol=1E-5, disp=root_disp)
    topt, rnfo = scipy.optimize.toms748(lambda t: froot.eval(t), tmin, tmax, xtol=1E-3, full_output=True, disp=True)
    wopt = froot.lastOptimalPoint
    tend        = time.time() - tend
    if( disp > 0 ):
        print('  Summary');
        print('      Iterations:          {:d}'.format(rnfo.iterations) )
        print('      Function calls:      {:d}'.format(rnfo.function_calls) )
        print('      Optimal value:       {:5.3E}'.format(topt) )
        print('      Elapsed time:        {:8.4f} seconds'.format(tend) )
    return wopt, la.norm( wopt.ravel(), 1 )

def CSRecoveryDebiasing(y, A, x, maxItns=1E4, dwAbsTol=1E-5, dfwAbsTol=1E-6, xthr=1E-6):
    # Find lipschitz constant
    smax        = OperatorNorm( A )
    L           = 1.01 * math.pow(smax, 2)
    # Find support of the input signal
    xsupp       = np.where( np.absolute(x) > xthr, 1, 0)
    # Initialize variables
    w           = x
    dwNrm       = np.inf
    itn         = 0
    fw          = 0.5 * math.pow(la.norm((y - A.eval(w, 1)).ravel(), 2), 2)
    # Optimization loop
    while( ( dwNrm > dwAbsTol or np.abs(dfw) > dfwAbsTol ) and itn <= maxItns):
        itn         = itn + 1
        wk          = w - (1/L) * A.eval(A.eval(w, 1) - y, 2)
        wk          = np.where(xsupp > 0, wk, 0)
        fwk         = 0.5 * math.pow(la.norm((y - A.eval(wk, 1)).ravel(), 2), 2)
        dfw         = fwk - fw
        dwNrm       = la.norm( (wk - w).ravel(), 2 )
        w           = wk
        fw          = fwk
    return w, fw

############################
## SAMPLING FUNCTIONS ######
###########################

def next_prime():
    def is_prime(num):
    #"Checks if num is a prime value"
        for i in range(2,int(num**0.5)+1):
            if(num % i)==0: return False
            return True
    prime = 3
    while(1):
        if is_prime(prime):
            yield prime
        prime += 2
def vdc(n, base=2):
    vdc, denom = 0, 1
    while n:
        denom *= base
        n, remainder = divmod(n, base)
        vdc += remainder/float(denom)
    return vdc
def halton_sequence(size, dim):
    seq = []
    primeGen = next_prime()
    next(primeGen)
    for d in range(dim):
        base = next(primeGen)
        seq.append([vdc(i, base) for i in range(size)])
    return seq

def VardensSampling(shape, f):
    x = np.linspace(-1, 1, num=shape[1])
    y = np.linspace(-1, 1, num=shape[0])
    omega = np.full(shape, False)
    for Ix in range(0, shape[1]):
        for Iy in range(0, shape[0]):
            u = np.random.uniform(0, 1)
            if( u < f([ x[Ix], y[Iy] ]) ):
                omega[Ix, Iy] = True
    return np.fft.fftshift(omega)

def VardensTriangleSampling(shape, delta):
    if( delta < 0.25 ):
        c = delta / 0.25
        a = 0
    else:
        c = 1
        a = 2 * np.sqrt(delta) - 1
    x = np.linspace(-1, 1, num=shape[1])
    y = np.linspace(-1, 1, num=shape[0])
    omega = np.full(shape, False)
    for Ix in range(0, shape[1]):
        for Iy in range(0, shape[0]):
            if( np.max([ np.abs(x[Ix]), np.abs(y[Iy]) ]) < a ):
                omega[Ix, Iy] = True
            else:
                p = c
                if( np.abs(x[Ix]) > a ):
                    p = p * (1 - (np.abs(x[Ix]) - a) / (1 - a))
                if( np.abs(y[Iy]) > a ):
                    p = p * (1 - (np.abs(y[Iy]) - a) / (1 - a))
                u = np.random.uniform(0, 1)
                if( u < p ):
                    omega[Ix, Iy] = True
    return np.fft.fftshift(omega)

def VardensGaussianSampling(shape, delta):
    c = 2 * np.sqrt(delta / np.pi)
    s, rnfo = scipy.optimize.toms748(lambda t: scipy.special.erf(t) - c * t, 1E-6, 1/c, xtol=1E-3, full_output=True, disp=True)
    s = 1 / (s * np.sqrt(2))
    x = np.linspace(-1, 1, num=shape[1])
    y = np.linspace(-1, 1, num=shape[0])
    omega = np.full(shape, False)
    for Ix in range(0, shape[1]):
        for Iy in range(0, shape[0]):
            p = np.exp( -(x[Ix]**2 + y[Iy]**2) / (2 * s**2) )
            u = np.random.uniform(0, 1)
            if( u < p ):
                omega[Ix, Iy] = True
    return np.fft.fftshift(omega)

def VardensExponentialSampling(shape, delta):
    c = np.sqrt(delta)
    a, rnfo = scipy.optimize.toms748(lambda t: np.exp(-t) - 1 + c * t, 1E-6, 1/c, xtol=1E-3, full_output=True, disp=True)
    x = np.linspace(-1, 1, num=shape[1])
    y = np.linspace(-1, 1, num=shape[0])
    omega = np.full(shape, False)
    for Ix in range(0, shape[1]):
        for Iy in range(0, shape[0]):
            p = np.exp( -a * (np.abs(x[Ix]) + np.abs(y[Iy])) )
            u = np.random.uniform(0, 1)
            if( u < p ):
                omega[Ix, Iy] = True
    return np.fft.fftshift(omega)   

def bisect(f, a, b, tol=1E-3):
  #bisect (binary search) for function only defined for integers
  while f(a) < f(b):
    mid = int((a+b)/2)
    midval = f(mid)
    if midval < 0-tol:
      a = mid + 1
    elif midval > 0+tol:
      b = mid
    else: #midval within (-tol, tol)
      return mid
  #root not found, but return closest pt anyway 
  if np.abs(f(a)) < np.abs(f(b)):
    return a
  return b

def HaltonSampling(shape, p):
  def getPoints(numPts):
    pts = np.transpose(np.asarray(halton_sequence(numPts, 2)))
    pts[:,0] = pts[:,0]*shape[0]
    pts[:,1] = pts[:,1]*shape[1]
    pts = pts.astype(int)
    return pts
  def getRatio(numPts):
    pts = getPoints(numPts)
    return 1.0*np.unique(pts, axis=0).shape[0]/(np.prod(shape))
  f = lambda x: getRatio(x) - p
  numPts = bisect(f, int(p*np.prod(shape)), 4*np.prod(shape)) #f: increasing fcn
  pts = getPoints(numPts)
  indices = np.zeros(shape)
  indices[pts[:,0], pts[:,1]] = 1
  return indices

def generateSamplingMask(imsz, p, saType='bernoulli', num_patterns=1, seed=1234321):
  # p is the undersampling ratio: what you don't sample
  if(p < 0.0)or(p > 1.0):
    print('ERROR: Invalid undersampling ratio delta in generateSamplingMask.')
    sys.exit(-1)
  elif(p == 0.0):
    return np.full(imsz, True, dtype=bool)
  else:
    np.random.seed(seed)
    mask = np.empty((num_patterns, ) + imsz, dtype=np.bool)
    for k in range(num_patterns):
      #to keep undersampling the same for each slice
      if saType=='bernoulli':
        indices = bernoulli.rvs(size=(imsz), p=(1-p))
      elif saType =='vardentri':
        indices = VardensTriangleSampling(imsz, (1-p))
      elif saType =='vardengauss': #gaussian density
        indices = VardensGaussianSampling(imsz, (1-p))
      elif saType == 'vardenexp': #exponential density
        indices = VardensExponentialSampling(imsz, (1-p))
      elif saType == 'halton': #halton sequence
        indices = HaltonSampling(imsz, (1-p))
      else:
        print('ERROR: Invalid sampling type')
        sys.exit(-1)
      mask[k] = np.ma.make_mask(indices)
      # Return the complement of the mask
    return mask



# Available operators
from mrina.maps import OperatorLinear
from mrina.maps import OperatorWaveletToFourier
# Sampling masks
from mrina.mri_utils import generateSamplingMask
# Image cropping to next power of 2
from mrina.mri_utils import crop
# l1-norm solvers
from mrina.solver_l1_norm import RecoveryL1NormNoisy
# Orthogonal matching pursuit solvers
from mrina.solver_omp import lsQR
from mrina.solver_omp import OMPRecovery

